import re


__all__ = ["adjust_atomic_symbol", "description_to_symbols"]


def adjust_atomic_symbol(x):
    """Makes sure atomic symbol is right-aligned and upper case (PFANT convention)."""
    assert isinstance(x, str)
    return "%2s" % (x.strip().upper())


# Formulas that can be recognized in PFANT molecule header information ("titulo")
# This exists for backward compatibility. New molecules should have its symbols informed separately
#
# **Careful** If you add entried, make sure everything is in uppercase
_BUILTIN_FORMULAS = {
     "MGH": ["MG", "H"],
     "C2": ["C", "C"],
     "CN": ["C", "N"],
     "CH": ["C", "H"],
     "13CH": ["A", "H"],
     "12C16O": ["C", "O"],
     "NH": ["N", "H"],
     "OH": ["O", "H"],
     "FEH": ["FE", "H"],
     "TIO": ["TI", "O"],
    "CO": ["C", "O"],
}


def description_to_symbols(descr):
    """Finds molecular formula in description text and returns separate symbols

    This routine is similar to pfantlib.f90:assign_symbols_by_formula()

    It is used when the symbols are not informed separately
    after the first "#" in the molecule 'titulo'

    Args:
        descr: e.g., "12C16O INFRARED  X 1 SIGMA+,   version 15/oct/98"

    Returns: list of two elements, e.g., ["MG", " H"]
    """

    descr = descr.upper()
    for formula, symbols in _BUILTIN_FORMULAS.items():
        # Searches for formula at start or after whitespace
        if re.search(r'(^|\W){}([^0-9A-Z]|$)'.format(formula), descr):
            return [adjust_atomic_symbol(x) for x in symbols]
    return None
    # raise RuntimeError("Could not find a valid formula in description '{}'".format(descr))



from . import a_WFileAbonds
from . import a_WFileAbXFwhm
from . import a_WFileMain
from . import a_WOptionsEditor
from . import a_XAtomLinesEditor
from . import a_XFileAbonds
from . import a_XFileAtoms
from . import a_XFileAtomsHistogram
from . import a_XFileMain
from . import a_XFileMolecules
from . import a_XMainAbonds
from . import a_XMolLinesEditor
from . import a_XMulti
from . import a_XPFANT
from . import a_XRunnableManager

from .a_WFileAbonds import *
from .a_WFileAbXFwhm import *
from .a_WFileMain import *
from .a_WOptionsEditor import *
from .a_XAtomLinesEditor import *
from .a_XFileAbonds import *
from .a_XFileAtoms import *
from .a_XFileAtomsHistogram import *
from .a_XFileMain import *
from .a_XFileMolecules import *
from .a_XMainAbonds import *
from .a_XMolLinesEditor import *
from .a_XMulti import *
from .a_XPFANT import *
from .a_XRunnableManager import *



"""Visualization classes for pyfant file types"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from mpl_toolkits.mplot3d import Axes3D  # yes, required (see below)
import pyfant as pf
import astroapi as aa
from astroapi import Vis
from pyfant.datatypes import *


class VisFileToH(Vis):
    """
    Plots hydrogen lines: each atmospheric layer is plotted as a y-axis-dislocated
    Spectrum in a 3D plot.
    """

    input_classes = (FileToH,)
    action = "Visualize hydrogen lines profiles"

    def _do_use(self, r):
        fig = plt.figure()
        mpl.rcParams['legend.fontsize'] = 10
        fig.canvas.set_window_title(self.title)  # requires the Axes3D module
        ax = fig.gca(projection='3d')
        x = np.concatenate((2 * r.lambdh[0] - r.lambdh[-2::-1], r.lambdh))
        _y = np.ones(len(x))
        for i in range(r.th.shape[1]):
            z = np.concatenate((r.th[-2::-1, i], r.th[:, i]))
            # ax.plot(x, _y * (i + 1), np.log10(z), label='a', color='k')
            ax.plot(x, _y * (i + 1), z, label='a', color='k')
        ax.set_xlabel('Wavelength ($\AA$)')
        ax.set_ylabel("Atmospheric layer #")
        # ax.set_zlabel('log10(Intensity)')
        # ax.set_zlabel('?')
        plt.tight_layout()
        plt.show()


# # Editors

class VisAtoms(Vis):
    """Opens the ated window."""
    input_classes = (FileAtoms,)
    action = "Edit using atomic lines editor"

    def _do_use(self, r):

        form = aa.keep_ref(pf.XFileAtoms(self.parent_form))
        form.load(r)
        form.show()


class VisMolecules(Vis):
    """Opens the mled window."""
    input_classes = (FileMolecules,)
    action = "Edit using molecular lines editor"

    def _do_use(self, r):
        form = aa.keep_ref(pf.XFileMolecules(self.parent_form))
        form.load(r)
        form.show()


class VisMain(Vis):
    """Opens the mained window."""
    input_classes = (FileMain,)
    action = "Edit using main configuration file editor"

    def _do_use(self, r):
        form = aa.keep_ref(pf.XFileMain(self.parent_form, r))
        form.show()


class VisAbonds(Vis):
    """Opens the abed window."""
    input_classes = (FileAbonds,)
    action = "Edit using abundances file editor"

    def _do_use(self, r):
        form = aa.keep_ref(pf.XFileAbonds(self.parent_form, r))
        form.show()



# #
# #
# # b
# # For OH, the format of branch (Br) in the lower-state quanta field is 2A1 to accommodate the total orbital angular
# # momentum N as well as J
# #
# #
# #
# # References:
# # [1] Rothman, Laurence S., et al. "The HITRAN 2004 molecular spectroscopic database."
# #     Journal of Quantitative Spectroscopy and Radiative Transfer 96.2 (2005): 139-204.
#
#
#
# __all__ = ["FileHitran", "Atom", "HitranLine"]
#
# import pyfant as pf
# import hapi
# from ..misc import *
# from ..errors import *
# import struct
# import logging
# import sys
# import numpy as np
#
# _logger = logging.getLogger(__name__)
# _logger.addHandler(logging.NullHandler())
#
#
# class FileHitran(DataFile):
#     """
#     PFANT Atomic Lines
#     """
#
#     default_filename = "Hitran.dat"
#     attrs = ["Hitran", "num_lines"]
#
#     @property
#     def num_lines(self):
#         ret = sum(map(len, self.Hitran))
#         return ret
#
#     # Properties that iterate through the HitranLine objects to mount vectors
#     @property
#     def lines(self):
#         """Merges all HitranLine objects. Returns a list."""
#         ret = []
#         for a in self.Hitran:
#             ret.extend(a.lines)
#         return ret
#     @property
#     def lambda_(self):
#         return np.hstack([np.array([x.lambda_ for x in a.lines]) for a in self.Hitran])
#     @property
#     def kiex(self):
#         return np.hstack([np.array([x.kiex for x in a.lines]) for a in self.Hitran])
#     @property
#     def algf(self):
#         return np.hstack([np.array([x.algf for x in a.lines]) for a in self.Hitran])
#     @property
#     def ch(self):
#         return np.hstack([np.array([x.ch for x in a.lines]) for a in self.Hitran])
#     @property
#     def gr(self):
#         return np.hstack([np.array([x.gr for x in a.lines]) for a in self.Hitran])
#     @property
#     def ge(self):
#         return np.hstack([np.array([x.ge for x in a.lines]) for a in self.Hitran])
#     @property
#     def zinf(self):
#         return np.hstack([np.array([x.zinf for x in a.lines]) for a in self.Hitran])
#     @property
#     def abondr(self):
#         return np.hstack([np.array([x.abondr for x in a.lines]) for a in self.Hitran])
#
#     def __len__(self):
#         """Length of FileHitran object is defined as number of elements."""
#         return len(self.Hitran)
#
#     def __init__(self):
#         DataFile.__init__(self)
#
#         # list of Atom objects
#         self.Hitran = []
#
#     def cut(self, llzero, llfin):
#         """Keeps only the lines with their llzero <= lambda_ <= llfin."""
#         for i in reversed(list(range(len(self)))):
#             atom = self.Hitran[i]
#             atom._cut(llzero, llfin)
#             if len(atom) == 0:
#                 del self.Hitran[i]
#
#     def filter(self, function):
#         """
#         Filters atomic lines for which function(line) is true.
#
#         Arguments:
#           function -- receives an HitranLine object as argument.
#            Example: lambda line: line.algf >= -7
#         """
#         for i in reversed(list(range(len(self)))):
#             atom = self.Hitran[i]
#             atom.lines = list(filter(function, atom.lines))
#             if len(atom) == 0:
#                 del self.Hitran[i]
#
#     filter_lines = filter
#
#     def filter_Hitran(self, function):
#         """
#         Filters Atom objects for which function(line) is true.
#
#         Arguments:
#           function -- receives an Atomic object as argument.
#            Example: lambda atom: atom.ioni <= 2
#         """
#         self.Hitran = list(filter(function, self.Hitran))
#
#     def remove_element(self, elem):
#         """Removes given element (any ionization level)."""
#         elem = adjust_atomic_symbol(elem)
#         for i in reversed(list(range(len(self)))):
#             atom = self.Hitran[i]
#             if atom.elem == elem:
#                 del self.Hitran[i]
#
#
#     def _do_load(self, filename):
#         """Clears internal lists and loads from file."""
#
#         with open(filename, "r") as h:
#
#             r = 0 # counts rows of file
#             edict = {}  # links atomic symbols with Atom objects created (key is atomic symbol)
#             try:
#                 while True:
#                     line = HitranLine()
#
#                     # (EE)(I) --whitespace-- (float) --ignored...--
#                     temp = str_vector(h)
#                     elem, s_ioni = temp[0][:-1], temp[0][-1]
#                     line.lambda_ = float(temp[1])
#                     elem = adjust_atomic_symbol(elem)
#                     key = elem+s_ioni  # will gb.py elements by this key
#                     if key in edict:
#                         a = edict[key]
#                     else:
#                         a = edict[key] = Atom()
#                         a.elem = elem
#                         a.ioni = int(s_ioni)
#                         self.Hitran.append(a)
#                     a.lines.append(line)
#                     r += 1
#
#                     [line.kiex, line.algf, line.ch, line.gr, line.ge, line.zinf, line.abondr, finrai] = \
#                         float_vector(h)
#                     r += 1
#                     # last element is end-of-file flag "finrai"
#                     if finrai == 1:
#                         break
#             except Exception as e:
#                 raise type(e)(("Error around %d%s row of file '%s'" %
#                                (r+1, ordinal_suffix(r+1), filename))+": "+str(e)).with_traceback(sys.exc_info()[2])
#
#     def _do_save_as(self, filename):
#         with open(filename, "w") as h:
#             n = len(self.Hitran)
#             for i, e in enumerate(self.Hitran):
#                 p = len(e)
#                 for j, a in enumerate(e.lines):
#                     finrai = 1 if i == n-1 and j == p-1 else 0
#                     write_lf(h, "%2s%1d %10.3f" % (e.elem, e.ioni, a.lambda_))
#                     # Writing floating-point numbers with %.7g format creates a more compact
#                     # file. In Fortran it is read with '*' format, so it understands as
#                     # long as there is a space between numbers.
#                     write_lf(h, "%.7g %.7g %.7g %.7g %.7g %.7g %.7g %1d" % \
#                              (a.kiex, a.algf, a.ch, a.gr, a.ge, a.zinf, a.abondr, finrai))
#                     # old way write_lf(h, "%8.3f %.7g %8.3f %8.3f %8.3f %6.1f %3.1f %1d" % \
#                     #     (a.kiex, a.algf, a.ch, a.gr, a.ge, a.zinf, a.abondr, finrai))



#!/usr/bin/env python3

"""Abundances file editor"""

import sys
from PyQt4.QtGui import *
import argparse
import pyfant as pf
import astroapi as aa
import logging


aa.logging_level = logging.INFO
aa.flag_log_file = True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=aa.SmartFormatter
    )
    parser.add_argument('fn', type=str, help='abundances file name', default='abonds.dat', nargs='?')
    args = parser.parse_args()

    m = pf.FileAbonds()
    m.load(args.fn)
    app = aa.get_QApplication([])
    form = pf.XFileAbonds()
    form.show()
    form.load(m)
    sys.exit(app.exec_())



#!/usr/bin/env python3
"""
Converts VALD3 atomic/molecular lines file to PFANT atomic lines file.

Molecular lines are skipped.

"""

import argparse
import logging
import numpy as np
import sys
import pyfant as pf
import astroapi as aa


aa.logging_level = logging.INFO
aa.flag_log_file = True


DEFOUT = "atoms-untuned-<fn_input>"

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=aa.SmartFormatter
     )
    parser.add_argument('fn_input', type=str, help='input file name', nargs=1)
    parser.add_argument('fn_output', type=str, help='output file name', nargs="?",
     default=DEFOUT)
    parser.add_argument('--min_algf', type=float, nargs='?', default=-7,
     help='minimum algf (log gf)')
    parser.add_argument('--max_kiex', type=float, nargs='?', default=15,
     help='maximum kiex')

    args = parser.parse_args()
    logger = aa.get_python_logger()

    fn_out = args.fn_output
    if fn_out == DEFOUT:
        fn_out = "atoms-untuned-"+args.fn_input[0]

    logger.info("Converting file...")
    with open(args.fn_input[0], 'r') as file_:
        file_atoms = pf.vald3_to_atoms(file_)

    n0 = file_atoms.num_lines
    logger.info("Number of lines before filtering: %d" % n0)

    logger.info("Removing all (algf < %g)..." % args.min_algf)
    file_atoms.filter(lambda line: line.algf >= args.min_algf)
    logger.info("Number of lines removed: %d" % (n0-file_atoms.num_lines))

    n0 = file_atoms.num_lines
    logger.info("Filtering all (kiex > %g)..." % args.max_kiex)
    file_atoms.filter(lambda line: line.kiex <= args.max_kiex)
    logger.info("Number of lines removed: %d" % (n0-file_atoms.num_lines))

    n0 = file_atoms.num_lines
    logger.info("Removing hydrogen lines...")
    file_atoms.remove_element("h")
    logger.info("Number of lines removed: %d" % (n0-file_atoms.num_lines))

    n0 = file_atoms.num_lines
    logger.info("Removing helium lines...")
    file_atoms.remove_element("he")
    logger.info("Number of lines removed: %d" % (n0-file_atoms.num_lines))

    n0 = file_atoms.num_lines
    logger.info("Removing F lines (no record in partit.dat)...")
    file_atoms.remove_element("F")
    logger.info("Number of lines removed: %d" % (n0-file_atoms.num_lines))

    n0 = file_atoms.num_lines
    logger.info("Removing Ne lines (no record in partit.dat)...")
    file_atoms.remove_element("Ne")
    logger.info("Number of lines removed: %d" % (n0-file_atoms.num_lines))

    logger.info("Number of lines after filtering: %d" % file_atoms.num_lines)

    logger.info("Saving file...")
    file_atoms.save_as(fn_out)
    print("File %s was successfully created." % fn_out)
    zz = file_atoms.zinf
    if len(zz) > 0:
        if np.all(zz == zz[0]):
            logger.warning("All zinf's set to %g. Now they must be fine-tuned with tune-zinf.py." % zz[0])



from PyQt4.QtCore import *
from PyQt4.QtGui import *
from .a_WDBMolecule import WDBMolecule
import astroapi as aa


__all__ = ["WMolConst"]


class WMolConst(aa.WBase):
    """
    Widget for the user to inform molecular parameters. Has a embedded WDBMolecule
    """

    # Emitted whenever the value of the molecule id changes
    # (actually propagates signal from the embedded WDBMolecule widget)
    id_changed = pyqtSignal()

    @property
    def fieldnames(self):
        return self._fieldnames

    @property
    def constants(self):
        """Returns a dictionary with all molecular constants"""
        ret = {}
        for fieldname in self._fieldnames:
            ret[fieldname] = self[fieldname]
        return ret

    @property
    def row(self):
        """Wraps WDBMolecule.row"""
        return self.w_mol.row

    def __init__(self, *args):
        aa.WBase.__init__(self, *args)

        self.flag_populating = False  # activated when populating table

        # # Internal state

        #
        self._fieldnames = ["fe", "do", "am", "bm", "ua", "ub", "te", "cro", "s"]


        # dictionary {(field name): (edit object), }
        # (will be populated later below together with edit widgets creation)
        self._edit_map = {}


        # # GUI design

        l = QVBoxLayout(self)
        l.setMargin(1)

        # ## State widget
        a = self.w_mol = WDBMolecule(self.parent_form)
        l.addWidget(a)
        a.id_changed.connect(self._w_mol_id_changed)

        # ## Title label
        a = self.apa = self.keep_ref(QLabel("<h4>Values to be used in calculations</h4>"))
        l.addWidget(a)
        a.setToolTip("If you values below, this will affect the conversion calculation, but the information will not be written back to the database")

        # ## Molecular constants edit fields

        lg = QGridLayout()
        l.addLayout(lg)
        nr, nc, n = 3, 3, len(self._fieldnames)
        # TODO not working but maybe wont use this widget anymore
        ti = aa.get_table_info("moldb", "molecule")
        for j in range(nc):
            # ### One grid layout for each column of fields
            ii = range(j, n, nc)
            for i in range(len(ii)):
                fieldname = self._fieldnames[ii[i]]
                info = ti[fieldname]
                caption = info["caption"] or fieldname
                a = QLabel(caption)
                e = QLineEdit("")
                tooltip = info["tooltip"]
                if tooltip:
                    a.setToolTip(tooltip)
                    e.setToolTip(tooltip)
                self._edit_map[fieldname] = e
                lg.addWidget(a, i, j*2)
                lg.addWidget(e, i, j*2+1)
            # if len(ii) < nr:
            #     # Adds spacer to cause the effect of aligning fields to the top when the column is not
            #     # full of fields
            #     lv.addItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding), i+1, 0, 1, 2)


    def __getitem__(self, fieldname):
        """Allows dict-like access of molecular constants of interest. Returns float value or None"""
        if fieldname in self._fieldnames:
            try:
                return float(self._edit_map[fieldname].text())
            except ValueError:
                return None


    def validate(self):
        """Returns True if all edit fields can be converted to float, otherwise False"""
        ret = True
        for fieldname in self._fieldnames:
            try:
                _ = float(self[fieldname])
            except ValueError:
                self.add_log_error("Failed converting field '{}' to float".format(fieldname))
                ret = False
                break
        return ret


    def None_to_zero(self):
        """Fills missing values with zeros"""
        for edit in self._edit_map.values():
            if len(edit.text().strip()) == 0:
                edit.setText("0")


    def _w_mol_id_changed(self):
        """Transfer values from self.data to edit fields below the table widget"""
        # id_ = print("State id changed to ", self.w_state.id)
        row = self.w_mol.row
        # if not row:
        #     return
        for fieldname, edit in self._edit_map.items():
            value = "" if not row else row[fieldname]
            edit.setText(str(value) if value is not None else "")

        self.id_changed.emit()



from .filedissoc import *
from .fileabonds import *
from .fileabsoru2 import *
from .fileabxfwhm import *
from .fileatoms import *
from .filehmap import *
from .filemain import *
from .filemolecules import *
from .fileoptions import *
from .filepartit import *
from .filetoh import *
from .filehitran import *
from .filevald3 import *
from .filemoldb import *
from .filehitrandb import *

from . import filedissoc
from . import fileabonds
from . import fileabsoru2
from . import fileabxfwhm
from . import fileatoms
from . import filehmap
from . import filemain
from . import filemolecules
from . import fileoptions
from . import filepartit
from . import filetoh
from . import filehitran
from . import filevald3



"""Represents SQLite database of molecular constants"""

import astroapi as aa
import pyfant as pf
import sqlite3
import tabulate


__all__ = ["FileMolDB"]

#
# __fileobj = None
# def get_conn():
#     return aa.get_conn(__ALIAS)


class FileMolDB(aa.FileSQLiteDB):
    description = "Database of Molecular Constants"
    default_filename = "moldb.sqlite"
    gui_data = {
        "molecule": {
            "name": [None, "name of molecule, <i>e.g.</i>, 'OH'"],
            "symbol_a": ["Symbol A", "symbol of first element"],
            "symbol_b": ["Symbol B", "symbol of second element"],
            "fe": [None, "oscillator strength"],
            "do": [None, "dissociation constant (eV)"],
            "am": [None, "mass of first element"],
            "bm": [None, "mass of second element"],
            "ua": [None, "value of partition function for first element"],
            "ub": [None, "value of partition function for second element"],
            "te": [None, "electronic term"],
            "cro": [None, "delta Kronecker (0: sigma transitions; 1: non-Sigma transitions)"],
            "s": [None, "?doc?"]
      },
        "state": {
    "omega_e": ["ω<sub>e</sub>", "vibrational constant – first term (cm<sup>-1</sup>)"],
    "omega_ex_e": ["ω<sub>e</sub>x<sub>e</sub>", "vibrational constant – second term (cm<sup>-1</sup>)"],
    "omega_ey_e": ["ω<sub>e</sub>y<sub>e</sub>", " vibrational constant – third term (cm<sup>-1</sup>)"],
    "B_e": ["B<sub>e</sub>", "rotational constant in equilibrium position (cm<sup>-1</sup>)"],
    "alpha_e": ["α<sub>e</sub>", "rotational constant – first term (cm<sup>-1</sup>)"],
    "D_e": ["D<sub>e</sub>", "centrifugal distortion constant (cm<sup>-1</sup>)"],
    "beta_e": ["β<sub>e</sub>", "rotational constant – first term, centrifugal force (cm<sup>-1</sup>)"],
      }
    }

    def _create_schema(self):
        conn = self.get_conn()
        c = conn.cursor()
        c.execute("""create table molecule (id integer primary key,
                                            formula text unique,
                                            name text,
                                            symbol_a text,
                                            symbol_b text,
                                            fe real,
                                            do real,
                                            am real,
                                            bm real,
                                            ua real,
                                            ub real,
                                            te real,
                                            cro real,
                                            s real
                                           )""")
        # Note that it has no primary key
        c.execute("""create table state (id integer primary key,
                                         id_molecule integer,
                                         State text,
                                         T_e real,
                                         omega_e real,
                                         omega_ex_e real,
                                         omega_ey_e real,
                                         B_e real,
                                         alpha_e real,
                                         gamma_e real,
                                         D_e real,
                                         beta_e real,
                                         r_e real,
                                         Trans text,
                                         nu_00 real,
                                         s_label text,
                                         s_multiplicity integer,
                                         s_spdf integer,
                                         s_parity text,

                                        )""")

        conn.commit()
        conn.close()


    def _populate(self):
        """Tries to download data from NIST Web Book online"""

        conn = self.get_conn()
        # TODO gather more formulae
        formulae = ["MgH", "C2", "CN", "CH", "NH", "CO", "OH", "FeH", "TiO"]

        # Uses PFANT/data/common/molecules.dat to retrieve "fe", "do", "am", etc.
        filemol = pf.FileMolecules()
        filemol.load(pf.get_pfant_data_path("common", "molecules.dat"))
        bysym = dict([(tuple(m.symbols), m) for m in filemol])


        for formula in formulae:
            try:
                data, _, name = pf.get_nist_webbook_constants(formula)

                fe, do, am, bm, ua, ub, te, cro, s = None, None, None, None, None, None, None, None, \
                                                     None

                # Tries to retrieve "fe", "do" etc from molecules.dat
                symbols = pf.description_to_symbols(formula)
                if not symbols:
                    aa.get_python_logger().warning("Formula '{}' not in internal BUILTIN_FORMULAS, "
                        "and probably this molecule is also not in PFANT/data/common/molecules.dat".
                        format(formula))

                    symbols = ["", ""]
                else:
                    m = bysym.get(tuple(symbols))
                    if m:
                        fe = m.fe
                        do = m.do
                        am = m.am
                        bm = m.bm
                        ua = m.ua
                        ub = m.ub
                        te = m.te
                        cro = m.cro
                        s = m.s

                symbols = [x.strip() for x in symbols]
                conn.execute("insert into molecule "
                             "(formula, name, symbol_a, symbol_b, fe, do, am, bm, ua, ub, te, cro, s) "
                             "values (?,?,?,?,?,?,?,?,?,?,?,?,?)",
                             (formula, name, symbols[0], symbols[1], fe, do, am, bm, ua, ub, te, cro, s))

                id_molecule = conn.execute("select last_insert_rowid() as id").fetchone()["id"]
                for state in data:
                    # **Note** assumes that the columns in data match the
                    # (number of columns in the state table - 2) and their order
                    conn.execute("insert into state values (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)",
                                 [None, id_molecule]+state)

                conn.commit()

            except:
                aa.get_python_logger().exception("Failed for molecule '{}'".format(formula))

        conn.close()


    def query_molecule(self, **kwargs):
        """Convenience function to query 'molecule' table

        Args:
            **kwargs: filter fieldname=value pairs to be passed to WHERE clause

        Returns: sqlite3 cursor
        """
        where = ""
        if len(kwargs) > 0:
            where = " where " + " and ".join([key + " = ?" for key in kwargs])
        conn = self.get_conn()
        sql = """select * from molecule{} order by formula""".format(where)
        r = conn.execute(sql, list(kwargs.values()))
        return r


    def query_state(self, **kwargs):
        """Convenience function to query 'state' table

        Args, Returns: see query_molecule
        """
        where = ""
        if len(kwargs) > 0:
            where = " where " + " and ".join([key + " = ?" for key in kwargs])
        conn = self.get_conn()
        sql = """select molecule.formula, state.* from state
                 join molecule on state.id_molecule = molecule.id{}""".format(where)
        r = conn.execute(sql, list(kwargs.values()))
        return r


    def test_query_state(self):
        """
        Test function

        >>> f = FileMolDB()
        >>> f.init_default()
        >>> conn = f.get_conn()
        >>> cursor = conn.execute("select * from state where id = 10")
        >>> row = cursor.fetchone()
        >>> print(row)
        MyDBRow([('id', 10), ('id_molecule', 2), ('State', 'F ⁱPi_u'), ('T_e', 75456.9), ('omega_e', 1557.5), ('omega_ex_e', None), ('omega_ey_e', None), ('B_e', 1.645), ('alpha_e', 0.019), ('gamma_e', None), ('D_e', 6e-06), ('beta_e', None), ('r_e', 1.307), ('Trans', 'F ← X R'), ('nu_00', 74532.9)])
        """


    def print_states(self, **kwargs):
        """
        Prints states table in console

        Args:
            **kwargs: arguments passed to query_state()

        >>> f = FileMolDB()
        >>> f.init_default()
        >>> f.print_states(formula="OH")
        """
        r = self.query_state(**kwargs)
        data, header0 = aa.cursor_to_data_header(r)

        # ti = aa.get_table_info(_ALIAS, "state")
        # header = [(ti[name]["caption"] or name) if ti.get(name) else name for name in header0]

        header = header0

        print(tabulate.tabulate(data, header))



__all__ = ["branch_to_Jl", "global_quanta_to_branch"]

_J2L_MAP = {"P": -1, "Q": 0, "R": 1}


def branch_to_Jl(Br, J2l):
    """Returns J superior given J inferior and the branch"""
    return J2l + _J2L_MAP[Br]


# TODO there is more ask BLB. "pois ha os dubletos que teem 6 ramos e tripletos teem 9 ramos"
def global_quanta_to_branch(Jl, J2l):
    """

    Args:
        Jl: J upper or J'
        J2l: J lower or  J''

    Returns: branch letter: "P"/"Q"/"R"
    """
    if Jl > J2l:
        return "R"
    if Jl == J2l:
        return "Q"
    return "P"



"""Represents SQLite database of molecular constants"""

import astroapi as aa
import pyfant as pf
import tabulate


__all__ = ["FileHitranDB"]

#
# __fileobj = None
# def get_conn():
#     return aa.get_conn(__ALIAS)


class FileHitranDB(aa.FileSQLiteDB):
    description = "HITRAN Molecules Catalogue"
    default_filename = "hitrandb.sqlite"
    gui_data = {
        "molecule": {
            "name": [None, "name of molecule, <i>e.g.</i>, 'OH'"],
            "symbol_a": ["Symbol A", "symbol of first element"],
            "symbol_b": ["Symbol B", "symbol of second element"],
            "fe": [None, "oscillator strength"],
            "do": [None, "dissociation constant (eV)"],
            "am": [None, "mass of first element"],
            "bm": [None, "mass of second element"],
            "ua": [None, "value of partition function for first element"],
            "ub": [None, "value of partition function for second element"],
            "te": [None, "electronic term"],
            "cro": [None, "delta Kronecker (0: sigma transitions; 1: non-Sigma transitions)"],
            "s": [None, "?doc?"]
      },
        "state": {
    "omega_e": ["ω<sub>e</sub>", "vibrational constant – first term (cm<sup>-1</sup>)"],
    "omega_ex_e": ["ω<sub>e</sub>x<sub>e</sub>", "vibrational constant – second term (cm<sup>-1</sup>)"],
    "omega_ey_e": ["ω<sub>e</sub>y<sub>e</sub>", " vibrational constant – third term (cm<sup>-1</sup>)"],
    "B_e": ["B<sub>e</sub>", "rotational constant in equilibrium position (cm<sup>-1</sup>)"],
    "alpha_e": ["α<sub>e</sub>", "rotational constant – first term (cm<sup>-1</sup>)"],
    "D_e": ["D<sub>e</sub>", "centrifugal distortion constant (cm<sup>-1</sup>)"],
    "beta_e": ["β<sub>e</sub>", "rotational constant – first term, centrifugal force (cm<sup>-1</sup>)"],
      }
    }

    def _create_schema(self):
        conn = self.get_conn()
        c = conn.cursor()







        c.execute("""create table molecule (ID integer unique, Formula text unique, Name text)""")
        # **Note** isotopologue.ID is not unique, it starts over for each molecule
        c.execute("""create table isotopologue (ID integer,
                                                ID_molecule integer,
                                                Formula text unique,
                                                AFGL_Code integer,
                                                Abundance real
                                               )""")
        conn.commit()
        conn.close()

    def _populate(self):
        conn = self.get_conn()
        try:

            mols, _ = pf.convmol.get_hitran_molecules()

            conn.executemany("insert into molecule values (?,?,?)", mols)

            aa.get_python_logger().info("Inserted {} molecules".format(len(mols)))

            for mol in mols:
                isos, _ = pf.convmol.get_hitran_isotopologues(mol[0])

                for iso in isos:
                    try:
                        conn.execute("insert into isotopologue values(?,?,?,?,?)",
                                     [iso[0], mol[0]] + iso[1:])
                    except:
                        aa.get_python_logger().exception("Tried to insert this: {}".format(iso))
                        raise

                aa.get_python_logger().info("Inserted {} isotopologues for molecule '{}' ({})".
                                            format(len(isos), *mol[1:3]))
        finally:
            conn.commit()
            conn.close()

        conn.close()


    def query_molecule(self, **kwargs):
        """Convenience function to query 'molecule' table

        Args:
            **kwargs: filter fieldname=value pairs to be passed to WHERE clause

        Returns: sqlite3 cursor
        """
        where = ""
        if len(kwargs) > 0:
            where = " where " + " and ".join([key + " = ?" for key in kwargs])
        conn = self.get_conn()
        sql = """select * from molecule{} order by ID""".format(where)
        r = conn.execute(sql, list(kwargs.values()))
        return r

    def query_isotopologue(self, **kwargs):
        """Convenience function to query 'isotopologue' table

        Args, Returns: see query_molecule

        Example:
        >>> f = FileHitranDB()
        >>> f.init_default()
        >>> _ = f.query_isotopologue(**{"molecule.formula": "OH"})
        """
        where = ""
        if len(kwargs) > 0:
            where = " where " + " and ".join([key + " = ?" for key in kwargs])
        conn = self.get_conn()
        sql = """select molecule.formula as m_formula, isotopologue.* from isotopologue
                 join molecule on isotopologue.id_molecule = molecule.id{}""".format(where)
        r = conn.execute(sql, list(kwargs.values()))
        return r

    def print_isotopologues(self, **kwargs):
        """
        Prints isotopologues table in console

        Args:
            **kwargs: arguments passed to query_state()

        >>> f = FileHitranDB()
        >>> f.init_default()
        >>> f.print_isotopologues(**{"molecule.formula": "CO"})
        """
        r = self.query_isotopologue(**kwargs)
        data, header = aa.cursor_to_data_header(r)
        print(tabulate.tabulate(data, header))



"""VALD3 atomic or molecular lines file"""


__all__ = ["Vald3Species", "FileVald3", "Vald3Line"]

# from ..liblib import *
import sys
import astroapi as aa
import io


@aa.froze_it
class Vald3Species(aa.AttrsPart):
    """
    Represents species of VALD3 **extended format** atomic/molecular lines file

    Args:
        formula: e.g., "OH", "Fe"
        ioni: ionization level, >= 1


    Examples: "OH 1", "Fe 2"

    """
    attrs = ["formula", "ioni"]

    # # Properties that iterate through the Vald3Line objects to mount vectors
    # @property
    # def lambda_(self):
    #     return np.array([x.lambda_ for x in self.lines])

    def __init__(self, formula=None, ioni=None):
        aa.AttrsPart.__init__(self)
        self.lines = []  # list of Vald3Line
        self.formula = None
        self.ioni = None

    def __len__(self):
        return len(self.lines)

    def __str__(self):
        return "{} {}".format(self.formula, self.ioni)

    def __repr__(self):
        return "'%s%s'" % (self.formula, self.ioni)

    def __iter__(self):
        return iter(self.lines)

    def _cut(self, llzero, llfin):
        """Keeps only the lines with their llzero <= lambda_ <= llfin."""
        for i in reversed(list(range(len(self)))):
            if not (llzero <= self.lines[i].lambda_ <= llfin):
                del self.lines[i]


@aa.froze_it
class Vald3Line(aa.AttrsPart):
    attrs = ["lambda_", "loggf", "Jl", "J2l", "vl", "v2l"]

    def __init__(self):
        aa.AttrsPart.__init__(self)

        self.lambda_ = None
        self.loggf = None
        self.Jl = None
        self.J2l = None
        self.vl = None
        self.v2l = None


class FileVald3(aa.DataFile):
    """
    VALD3 atomic or molecular lines file

    **Note** Load only

    **Note** Does not load all file fields, only those needed to convert this file to
             PFANT molecular lines format (20161216)
    """

    attrs = ["speciess", "num_lines"]

    @property
    def num_lines(self):
        ret = sum(map(len, self.speciess))
        return ret

    def __len__(self):
        return len(self.speciess)

    def __init__(self):
        aa.DataFile.__init__(self)

        # list of Atom objects
        self.speciess = []

    def __iter__(self):
        return iter(self.speciess)

    def remove_formula(self, formula):
        """Removes given element (any ionization level)."""
        formula = aa.adjust_atomic_symbol(formula)
        for i in reversed(list(range(len(self)))):
            atom = self.speciess[i]
            if atom.formula == formula:
                del self.speciess[i]


    def _do_load(self, filename):
        """Clears internal lists and loads from file."""

        with open(filename, "r") as h:
            self._do_load_h(h, filename)

    def _do_load_h(self, h, filename):
        # ---SAMPLE FILE---
        #                                                                    Lande factors      Damping parameters
        # Elm Ion      WL_air(A)   log gf* E_low(eV) J lo  E_up(eV) J up  lower  upper   mean   Rad.  Stark  Waals
        # 'OH 1',       16401.409,  -8.587,  4.0602, 39.5,  4.8160, 39.5,99.000,99.000,99.000, 0.000, 0.000, 0.000,
        # '  Hb                                                     2s2.3s2.1p3          X,2,1,f,40,3'
        # '  Hb                                                 2s2.3s2.1p3              X,2,1,f,39,7'

        r = 0  # counts rows of file
        speciess = {}
        try:
            # Uses header as "magic characters"
            s = h.readline().strip("\n")
            if s != "                                                                   Lande factors      Damping parameters":
                raise RuntimeError("This doesn't appear to be a VALD3 extended-format atomic/molecular lines file")
            r += 1
            h.readline()
            r += 1

            while True:
                line = Vald3Line()

                s = h.readline().strip()
                if len(s) == 0 or not s.startswith("'"):
                    # EOF: blank line or references section
                    break

                fields = s.split(",")
                formula, s_ioni = fields[0][1:-1].split(" ")
                line.lambda_ = float(fields[1])
                line.loggf = float(fields[2])
                line.Jl = float(fields[6])  # J_up
                line.J2l = float(fields[4])  # J_lo
                r += 1


                if formula in aa.symbols:
                    # Skips energy levels information for the atoms
                    for _ in range(2):
                        h.readline()
                        r += 1
                else:
                    s = h.readline().strip()
                    line.vl = float(s[s.rfind(",") + 1:-1])  # v superior / v upper
                    r += 1
                    s = h.readline().strip()
                    line.v2l = float(s[s.rfind(",") + 1:-1])  # v inferior / v lower
                    r += 1

                key = formula + s_ioni  # will gb.py elements by this key
                if key in speciess:
                    a = speciess[key]
                else:
                    a = speciess[key] = Vald3Species()
                    a.formula = formula
                    a.ioni = int(s_ioni)
                    self.speciess.append(a)
                a.lines.append(line)

                # Skips the citation line
                h.readline()
                r += 1

        except Exception as e:
            raise type(e)(("Error around %d%s row of file '%s'" %
                           (r + 1, aa.ordinal_suffix(r + 1), filename)) + ": " + str(
                e)).with_traceback(sys.exc_info()[2])


def _fake_file():
    """Returns StringIO file to test FileVald3 class"""

    f = io.StringIO()
    f.write("""                                                                   Lande factors      Damping parameters
Elm Ion      WL_air(A)   log gf* E_low(eV) J lo  E_up(eV) J up  lower  upper   mean   Rad.  Stark  Waals
'Fe 1',       16400.045,  -4.131,  6.5922,  6.0,  7.3480,  5.0, 1.320, 1.090, 1.910, 8.140,-3.840,-7.330,
'  LS                                                                       3d7.(4F).4d f5G'
'  JK                                                             3d7.(4F<7/2>).4f 2[11/2]*'
'K14               Kurucz Fe I 2014   1 K14       1 K14       1 K14       1 K14       1 K14       1 K14       1 K14       1 K14       1 K14     Fe            '
'CO 1',       16400.058,  -5.951,  1.3820, 38.0,  2.1378, 39.0,99.000,99.000,99.000, 2.000, 0.000, 0.000,
'  Hb                                                4s2.5s2.1p4              X,2,0,,none,4'
'  Hb                                                    4s2.5s2.1p4          X,2,0,,none,7'
'KCO                                  2 KCO       2 KCO       2 KCO       2 KCO       2 KCO       2 KCO       2 KCO       2 KCO       2 KCO     (12)C(16)O    '
'CO 1',       16400.090,  -8.995,  1.1482, 53.0,  1.9040, 54.0,99.000,99.000,99.000, 2.000, 0.000, 0.000,
'  Hb                                                4s2.5s2.1p4              X,2,0,,none,2'
'  Hb                                                    4s2.5s2.1p4          X,2,0,,none,5'
'KCO                                  2 KCO       2 KCO       2 KCO       2 KCO       2 KCO       2 KCO       2 KCO       2 KCO       2 KCO     (12)C(18)O    '
""")

    f.seek(0)
    return f

def _test():
    """Test code in docstring

    >>> h = _fake_file()
    >>> f = FileVald3()
    >>> f._do_load_h(h, "_fake_file")
    """
    return

# h = _fake_file()
# f = FileVald3()
# f._do_load_h(h, "_fake_file")



# # Temporary imports
#   =================
# These modules should be be del'eted at the end
import astroapi as aa


# # Setup
#   =====
SESSION_PREFIX_SINGULAR = 'session-'
SESSION_PREFIX_PLURAL = 'session-'
MULTISESSION_PREFIX = 'multi-session-'

# ## Colors definition
# Color for labels indicating a star parameter
COLOR_STAR = "#2A8000"
# Color for labels indicating a software configuration parameter
COLOR_CONFIG = "#BD6909"


# # Imports
#   =======
from .errors import *
from .liblib import *
from .datatypes import *
from .conf import *
from .runnables import *
from .rm import *
from .vis import *
from .util import *
from .from_vald import *
from .multirunnable import *
from .gui import *
from .paths import *
from . import datatypes
from . import liblib
from . import gui
from . import vis
from . import convmol

# # Function to access package-specific config file
#   ===============================================
def get_config():
    """Returns PyfantConfigObj object that corresponds to file ~/.pyfant.conf"""
    return aa.get_config_obj(".pyfant.conf")


# # # Function to be called from astroapi package
# #   ==========================================
# # **Note** To make astroapi execute this, astroapi/__init__.py has to be changed
# def _setup_astroapi():
#     """Adds entries to the astroapi module"""
#
#     aa._classes_txt.extend([FileAbsoru2, FileHmap, FileMain, FileDissoc,
#                            FileToH, FileAbonds, FileAtoms, FileMolecules])
#     aa._classes_file.extend(aa.get_classes_in_module(datatypes, aa.DataFile))
#     aa._classes_vis.extend(aa.get_classes_in_module(vis, aa.Vis))
#

# # Finally, gets rid of unwanted symbols in the workspace
#   ======================================================
del aa




"""
Conversion of molecular lines files.
"""

from .convlog import *
from .nistbot import *
from .hitranbot import *
from .calc_qgbd import *
from .conv_hitran import *
from .conv_vald3 import *
from .make_mol import *
from .a_WDBState import *
from .a_WDBMolecule import *
from .a_WMolConst import *
from .a_WStateConst import *
from .a_XConvMol import *
from .a_XFileMolDB import *



__all__ = ["FileMain"]
from ..errors import *
from astroapi import write_lf, DataFile, float_vector, str_vector, str2bool, bool2str

NINE_HUNDRED = 900

class FileMain(DataFile):
  """
  PFANT Stellar Main Configuration
  
  Imitates the logic of reader_main.f90::read_main().
  
  Attributes match reader_main.f90::main_* (minus the "main_" prefix)
  """

  description = "main configuration"
  default_filename = "main.dat"
  attrs = ["titrav", "pas", "fwhm", "ivtot", "vvt", "tolv",
           "teff", "glog", "asalog", "nhe", "ptdisk", "mu",
           "flprefix", "llzero", "llfin", "aint"]

  def __init__(self):
    DataFile.__init__(self)
    self.titrav = None
    self.ecrit = None
    self.pas = None
    self.echx = None
    self.echy = None
    self.fwhm = None
    self.ivtot = None
    self.vvt = None
    self.tolv = None
    self.teff = None
    self.glog = None
    self.asalog = None
    self.nhe = None
    self.inum = None
    self.ptdisk = None
    self.mu = None
    self.afstar = None
    self.xxcor = None
    self.flprefix = None
    self.llzero = None
    self.llfin = None
    self.aint = None
    self.filetohy = None

  def _do_load(self, filename):
    """Loads from file."""
  
    with open(filename, "r") as h:
      self.titrav = h.readline().strip()
      self.ecrit, self.pas, self.echx, self.echy, self.fwhm = str_vector(h)
  
      vvt = float(h.readline())
      self.ivtot = 1
      if vvt <= NINE_HUNDRED:
        self.vvt = [vvt]
      else:
        _ivtot = int(h.readline())
        self.tolv = float_vector(h)
        self.vvt = float_vector(h)
        self.ivtot = len(self.vvt)
  
        if not (self.ivtot == len(self.tolv) == len(vvt)):
          raise FileConsistencyError("ivtot must match len(tolv) must match len(vvt)")
  
      self.teff, self.glog, self.asalog, self.nhe, self.inum = str_vector(h)
      self.ptdisk, self.mu = str_vector(h)
      self.afstar = float(h.readline())
      self.xxcor = float_vector(h)
      self.flprefix = h.readline().strip()
      self.llzero, self.llfin, self.aint = float_vector(h)
  
      self.filetohy = []
      for s in h.readlines():
        s = s.strip()
        if len(s) > 0:
          self.filetohy.append(s)
  
    # remaining conversions
    self.ecrit, self.ptdisk = list(map(str2bool, (self.ecrit, self.ptdisk)))
    self.pas, self.echx, self.echy, self.fwhm, self.teff, self.glog, \
    self.asalog, self.nhe, self.mu = \
      list(map(float, (self.pas, self.echx, self.echy, self.fwhm, self.teff,
            self.glog, self.asalog, self.nhe, self.mu)))
    self.inum = int(self.inum)
  
  def _do_save_as(self, filename):
    """Saves to file."""
    assert isinstance(self.vvt, list), "vvt must be list!"
    list2str = lambda l: " ".join([str(x) for x in l])
    with open(filename, "w") as h:
      write_lf(h, "%s" % self.titrav)
      write_lf(h, "%s %s %s %s %s" % (bool2str(self.ecrit),
                      self.pas, self.echx, self.echy, self.fwhm))
      write_lf(h, "%s" % self.vvt[0])
      if self.vvt[0] > NINE_HUNDRED:
        write_lf(h, "%s" % len(self.vvt))  # self.ivtot)
        write_lf(list2str(self.tolv))
        write_lf(list2str(self.vvt))
  
      write_lf(h, list2str([self.teff, self.glog, self.asalog, self.nhe, self.inum]))
      write_lf(h, list2str([bool2str(self.ptdisk), self.mu]))
      write_lf(h, str(self.afstar))
      write_lf(h, list2str(self.xxcor))
      write_lf(h, self.flprefix)
      write_lf(h, list2str([self.llzero, self.llfin, self.aint]))
      for filetoh in self.filetohy:
        write_lf(h, filetoh)





from PyQt4.QtCore import *
from PyQt4.QtGui import *
import astroapi as aa
import pyfant as pf
# from a_WState import WState
# import moldb as db
from .a_WStateConst import WStateConst
from .a_WMolConst import WMolConst
from . import hapi
import os
import datetime


class _DataSource(aa.AttrsPart):
    """Represents a data source for molecular lines"""

    def __init__(self, name):
        aa.AttrsPart.__init__(self)
        self.name = name
        self.widget = None

    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, self.name)



# This defines the order of the panels
_NAMES = ["HITRAN", "VALD3", "TurboSpectrum", "Kurucz",]
_SOURCES = [_DataSource(x) for x in _NAMES]


class _WSource(aa.WBase):
    """Lists sources for molecular lines data"""


    @property
    def index(self):
        return self._get_index()

    @index.setter
    def index(self, x):
        self._buttons[x].setChecked(True)

    @property
    def source(self):
        """Returns _DataSource object or None"""
        i = self._get_index()
        if i < 0:
            return None
        return _SOURCES[i]

    index_changed = pyqtSignal()

    def __init__(self, *args):
        aa.WBase.__init__(self, *args)

        self._buttons = []
        self._last_index = -1

        lw = QVBoxLayout()
        self.setLayout(lw)

        for ds in _SOURCES:
            b = self.keep_ref(QRadioButton(ds.name))
            b.clicked.connect(self._button_clicked)
            self._buttons.append(b)
            lw.addWidget(b)

    def _button_clicked(self):
        i = self._get_index()
        if i != self._last_index:
            self._last_index = i
            self.index_changed.emit()


    def _get_index(self):
        for i, b in enumerate(self._buttons):
            assert isinstance(b, QRadioButton)
            if b.isChecked():
                return i
        return -1


class _WSelectSaveFile(aa.WBase):
    @property
    def value(self):
        return self._get_value()

    @value.setter
    def value(self, x):
        self.edit.setText(x)

    # # Emitted whenever the valu property changes **to a valid value**
    wants_auto = pyqtSignal()

    def __init__(self, *args):
        aa.WBase.__init__(self, *args)

        self._last_value = None

        self._type = None
        self.dialog_title = "Save output as"
        self.dialog_path = "."
        self.dialog_wild = "*.*;;*.dat"

        lw = QHBoxLayout()
        self.setLayout(lw)

        t = self.label = self.keep_ref(QLabel("Save output as"))
        lw.addWidget(t)

        e = self.edit = QLineEdit()
        e.textChanged.connect(self.edit_changed)
        t.setBuddy(e)
        lw.addWidget(e)
        # e.setReadOnly(True)

        b = self.button_auto = QToolButton()
        lw.addWidget(b)
        b.clicked.connect(self.wants_auto)
        b.setIcon(aa.get_icon("leaf-plant"))
        b.setToolTip("Make up file name")
        b.setFixedWidth(30)

        b = self.button = QToolButton()
        lw.addWidget(b)
        b.clicked.connect(self.on_button_clicked)
        b.setIcon(aa.get_icon("document-save"))
        b.setToolTip("Choose file name to save as")
        b.setFixedWidth(30)

        # Forces paint red if invalid
        self.edit_changed()

    def on_button_clicked(self, _):
        self._on_button_clicked()

    def edit_changed(self):
        flag_valid = self.validate()
        aa.style_widget_valid(self.edit, not flag_valid)
        # if flag_valid:
        #     self._wanna_emit()

    def validate(self):
        """Returns True/False whether value is valid, i.e., existing file or directory"""
        value = self._get_value().strip()
        return len(value) > 0 and not os.path.isdir(value)

    def _on_button_clicked(self):
        path_ = self.edit.text().strip()
        if len(path_) == 0:
            path_ = self.dialog_path
        res = QFileDialog.getSaveFileName(self, self.dialog_title, path_, self.dialog_wild)
        if res:
            self.edit.setText(res)
            self.dialog_path = res

    # def _wanna_emit(self):
    #     value_now = self._get_value()
    #     if value_now != self._last_value:
    #         self._last_value = value_now
    #         self.valueChanged.emit()

    def _get_value(self):
        return self.edit.text().strip()


class _WHitranPanel(aa.WBase):

    @property
    def data(self):
        """
        Returns value in hapi.LOCAL_TABLE_CACHE, or None. This variable is a dictionary

        See hapi.py for structure of this variable
        """
        idx = self.tableWidget.currentRow()
        if idx < 0:
            return None
        return hapi.LOCAL_TABLE_CACHE[self.tableWidget.item(idx, 0).text()]

    def __init__(self, *args):
        aa.WBase.__init__(self, *args)

        self._flag_populating = False

        lw = QVBoxLayout()
        self.setLayout(lw)

        lw.addWidget(self.keep_ref(QLabel(_SOURCES[0].name)))


        w = self.w_dir = aa.WSelectDir(self.parent_form)
        w.label.setText("HITRAN 'data cache' directory")
        w.valueChanged.connect(self.dir_changed)
        lw.addWidget(w)

        a = self.tableWidget = QTableWidget()
        lw.addWidget(a)
        a.setSelectionMode(QAbstractItemView.SingleSelection)
        a.setSelectionBehavior(QAbstractItemView.SelectRows)
        a.setEditTriggers(QTableWidget.NoEditTriggers)
        a.setFont(aa.MONO_FONT)
        a.setAlternatingRowColors(True)
        a.currentCellChanged.connect(self.on_tableWidget_currentCellChanged)

        # forces populate table with 'Python HITRAN API data cache' in local directory
        self.dir_changed()

    def on_tableWidget_currentCellChanged(self, curx, cury, prevx, prevy):
        pass

    def dir_changed(self):
        self._populate()


    def _populate(self):
        self._flag_populating = True
        try:
            # Changed HAPI working directory
            hapi.VARIABLES['BACKEND_DATABASE_NAME'] = self.w_dir.value
            # Loads all molecular lines data to memory
            hapi.loadCache()

            # Discounts "sampletab" table from HAPI cache, hence the "-1" below
            nr, nc = len(hapi.LOCAL_TABLE_CACHE)-1, 2
            t = self.tableWidget
            aa.reset_table_widget(t, nr, nc)
            t.setHorizontalHeaderLabels(["Table filename (.data & .header)", "Number of spectral lines"])

            i = 0
            for h, (name, data) in enumerate(hapi.LOCAL_TABLE_CACHE.items()):
                if name == "sampletab":
                    continue

                header = data["header"]

                item = QTableWidgetItem(name)
                t.setItem(i, 0, item)

                item = QTableWidgetItem(str(header["number_of_rows"]))
                t.setItem(i, 1, item)

                i += 1

            t.resizeColumnsToContents()

            # self._data = rows
            #
            # if restore_mode == "formula":
            #     if curr_row:
            #         self._find_formula(curr_row["formula"])
            # elif restore_mode == "index":
            #     if -1 < curr_idx < t.rowCount():
            #         t.setCurrentCell(curr_idx, 0)
        finally:
            self._flag_populating = False
            # self._wanna_emit_id_changed()


class _WVald3Panel(aa.WBase):
    """
    This panel allows to load a Vald3 file and browse through its species (molecules only)

    The goal is to choose one molecule
    """

    @property
    def data(self):
        """
        Returns FileVald3 with a single species"""

        idx = self.tableWidget.currentRow()
        if idx < 0:
            return None

        f = pf.FileVald3()
        f.speciess = [self._f.speciess[idx]]
        return f

    @property
    def is_molecule(self):
        data = self.data
        return data is not None and data.speciess[0].formula not in aa.symbols

    def __init__(self, *args):
        aa.WBase.__init__(self, *args)

        self._flag_populating = False
        self._f = None  # FileVald3

        lw = QVBoxLayout()
        self.setLayout(lw)

        lw.addWidget(self.keep_ref(QLabel(_SOURCES[0].name)))


        w = self.w_file = aa.WSelectFile(self.parent_form)
        w.label.setText("VALD3 file")
        w.valueChanged.connect(self.file_changed)
        lw.addWidget(w)

        a = self.tableWidget = QTableWidget()
        lw.addWidget(a)
        a.setSelectionMode(QAbstractItemView.SingleSelection)
        a.setSelectionBehavior(QAbstractItemView.SelectRows)
        a.setEditTriggers(QTableWidget.NoEditTriggers)
        a.setFont(aa.MONO_FONT)
        a.setAlternatingRowColors(True)
        a.currentCellChanged.connect(self.on_tableWidget_currentCellChanged)

        l = self.label_warning = QLabel()
        l.setStyleSheet("QLabel {{color: {}}}".format(aa.COLOR_WARNING))
        lw.addWidget(l)

        # forces populate table with 'Python HITRAN API data cache' in local directory
        # self.file_changed()

    def on_tableWidget_currentCellChanged(self, curx, cury, prevx, prevy):
        print("OOOOOOOOOOOOOOOOO")
        self.label_warning.setText("Need a molecule, not atom"
                                   if self.data is not None and not self.is_molecule else "")

    def file_changed(self):
        self._populate()


    def _populate(self):
        self._flag_populating = True
        try:
            f = self._f = pf.FileVald3()
            f.load(self.w_file.value)

            nr, nc = len(f), 3
            t = self.tableWidget
            aa.reset_table_widget(t, nr, nc)
            t.setHorizontalHeaderLabels(["VALD3 species", "Number of spectral lines", "Atom/Molecule"])

            for i, species in enumerate(f):
                item = QTableWidgetItem(str(species))
                t.setItem(i, 0, item)
                item = QTableWidgetItem(str(len(species)))
                t.setItem(i, 1, item)
                item = QTableWidgetItem("Atom" if species.formula in aa.symbols else "Molecule")
                t.setItem(i, 2, item)

            t.resizeColumnsToContents()

        except Exception as e:
            self._f = None
            self.add_log_error("Error reading contents of file '{}': '{}'".format(self.w_file.value, aa.str_exc(e)), True)
            raise

        finally:
            self._flag_populating = False
            # self._wanna_emit_id_changed()



class _WTurboSpectrumPanel(aa.WBase):
    def __init__(self, *args):
        aa.WBase.__init__(self, *args)

        lw = QVBoxLayout()
        self.setLayout(lw)

        lw.addWidget(self.keep_ref(QLabel(_SOURCES[1].name)))


class _WKuruczPanel(aa.WBase):
    def __init__(self, *args):
        aa.WBase.__init__(self, *args)

        lw = QVBoxLayout()
        self.setLayout(lw)

        lw.addWidget(self.keep_ref(QLabel(_SOURCES[2].name)))


class XConvMol(aa.XLogMainWindow):
    def __init__(self, *args):
        aa.XLogMainWindow.__init__(self, *args)

        tw0 = self.keep_ref(QTabWidget())
        self.setCentralWidget(tw0)

        # # First tab: polluted!!

        sp = self.keep_ref(QSplitter(Qt.Vertical))
        tw0.addTab(sp, "Molecular constants (Alt+&1)")

        # ## First widget of splitter
        w0 = self.keep_ref(QWidget())
        l0 = QVBoxLayout(w0)
        l0.setMargin(2)
        l0.setSpacing(2)

        a = self.title_mol = QLabel(aa.format_title0("Select a molecule:"))
        l0.addWidget(a)

        w = self.w_mol = WMolConst(self)
        w.layout().setContentsMargins(15, 1, 1, 1)
        w.id_changed.connect(self.mol_id_changed)
        l0.addWidget(w)


        # ## Second widget of splitter
        w1 = self.keep_ref(QWidget())
        l1 = QVBoxLayout(w1)
        l1.setMargin(2)
        l1.setSpacing(2)

        a = self.title_state = self.keep_ref(QLabel(aa.format_title0("States")))
        l1.addWidget(a)

        w = self.w_state = WStateConst(self)
        w.layout().setContentsMargins(15, 1, 1, 1)
        l1.addWidget(w)

        b = self.button_fill = QPushButton("Fill missing values with ZERO")
        b.clicked.connect(self.on_fill_missing)
        l1.addWidget(b)

        sp.addWidget(w0)
        sp.addWidget(w1)


        # # Second tab: files

        w = self.keep_ref(QWidget())
        tw0.addTab(w, "Conversion (Alt+&2)")


        # ## Vertical layout: source and destination stacked
        lsd = self.keep_ref(QVBoxLayout(w))

        # ### Horizontal layout: sources radio buttons, source-specific setup area

        lh = self.keep_ref(QHBoxLayout())
        lsd.addLayout(lh)

        # #### Vertical layout: source radio group box

        lss = QVBoxLayout()
        lh.addLayout(lss)

        # ##### Source radio buttons
        lss.addWidget(self.keep_ref(QLabel("<b>Source</b>")))
        w = self.w_source = _WSource(self)
        w.index_changed.connect(self.source_changed)
        lss.addWidget(w)
        lss.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))


        # #### Adds configuration panels for various sources
        # Only one panel should be visible at a time
        # **Note** The order here doesn't matter
        p0 = self.w_hitran = _WHitranPanel(self)
        p1 = self.w_vald3 = _WVald3Panel(self)
        p2 = self.w_turbo = _WTurboSpectrumPanel(self)
        p3 = self.w_kurucz = _WKuruczPanel(self)
        pp = [p0, p1, p2, p3]
        for ds, p in zip(_SOURCES, pp):
            ds.widget = p
            lh.addWidget(p)


        # tw1 = self.wt_source = QTabWidget()
        # lh.addWidget(tw1)
        # tw1.addTab(self.keep_ref(QLabel(_SOURCES[0].name)), _SOURCES[0].name)
        # tw1.addTab(self.keep_ref(QLabel(_SOURCES[1].name)), _SOURCES[1].name)
        # tw1.addTab(self.keep_ref(QLabel(_SOURCES[2].name)), _SOURCES[2].name)


        # ### Output file specification

        w0 = self.w_out = _WSelectSaveFile(self)
        w0.wants_auto.connect(self.wants_auto)
        lsd.addWidget(w0)

        # ### "Convert" button

        lmn = QHBoxLayout()
        lsd.addLayout(lmn)
        lmn.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))
        b = self.button_convert = QPushButton("&Run Conversion")
        b.clicked.connect(self.convert_clicked)
        lmn.addWidget(b)


        # # Log facilities

        sb = self.statusBar()
        l = self.label_last_log = QLabel()
        sb.addWidget(l)

        w = self.textEdit_log = QTextEdit()
        w.setReadOnly(True)
        tw0.addTab(w, "Log messages (Alt+&L)")

        # # Final adjustments

        tw0.setCurrentIndex(1)
        # Forces only one of the source panels to visible
        self.w_source.index = 0
        self.source_changed()
        self.setWindowTitle("(to) PFANT Molecular Lines Converter")

        aa.nerdify(self)


    def wants_auto(self):
        idx = self.w_source.index
        filename = None
        if idx == 0:
            lines = self.w_hitran.data
            if lines:
                filename = "{}.dat".format(lines["header"]["table_name"])

        if filename is None:
            # Default
            filename = aa.new_filename("mol", "dat")

        self.w_out.value = filename

    def on_fill_missing(self):
        self.w_mol.None_to_zero()
        self.w_state.None_to_zero()

    def source_changed(self):
        idx = self.w_source.index
        # print("SOURCE IS NOW ", _SOURCES[idx])
        for i, ds in enumerate(_SOURCES):
            pass
            ds.widget.setVisible(i == idx)
            # print("Widget", ds.widget, "is visible?", ds.widget.isVisible())

    def mol_id_changed(self):
        id_ = self.w_mol.w_mol.id
        row = self.w_mol.w_mol.row
        self.w_state.set_id_molecule(id_)
        s = "States (no molecule selected)" if not row else "Select a State for molecule '{}'".format(row["formula"])
        self.title_state.setText(aa.format_title0(s))

    def convert_clicked(self):
        cm = pf.convmol
        try:
            errors = []
            idx = self.w_source.index
            name = _SOURCES[idx].name
            mol_row = self.w_mol.row
            mol_consts = self.w_mol.constants
            state_consts = self.w_state.constants
            filename = self.w_out.value
            if not mol_row:
                errors.append("Molecule not selected")
            if any([x is None for x in mol_consts.values()]):
                errors.append("There are empty molecule-wide constants")
            if any([x is None for x in state_consts.values()]):
                errors.append("There are empty state-wide constants")
            if not self.w_out.validate():
                errors.append("Output filename is invalid")

            lines, sols_calculator = None, None
            if len(errors) == 0:
                # Source-dependant calculation of "sets of lines"
                if name == "HITRAN":
                    lines = self.w_hitran.data

                    if lines is None:
                        errors.append("HITRAN table not selected")
                    else:
                        sols_calculator = cm.hitran_to_sols
                elif name == "VALD3":
                    if not self.w_vald3.is_molecule:
                        errors.append("Need a VALD3 molecule")
                    else:
                        lines = self.w_vald3.data
                        sols_calculator = cm.vald3_to_sols
                else:
                    aa.show_message("{}-to-PFANT conversion not implemented yet, sorry".
                                    format(name))
                    return

            if len(errors) == 0:
                # Finally the conversion to PFANT molecular lines file

                # TODO **MAYBE NOT TIO-LIKE***
                mol_row.update(mol_consts)  # replaces possibly changed values for "fe", "do" etc.
                f, log = cm.make_file_molecules(mol_row, state_consts, lines,
                                                cm.calc_qgbd_tio_like, sols_calculator)
                ne = len(log.errors)
                if ne > 0:
                    self.add_log("Error messages:")
                    self.add_log("\n".join(log.errors))
                    self.add_log("{} error message{}".format(ne, "" if ne == 1 else "s"))

                if log.flag_ok:
                    f.save_as(filename)
                    self.add_log(
                        "Lines converted: {}/{}".format(f.num_lines, log.num_lines_in))

                    self.add_log("File '{}' generated successfully".format(filename))
                else:
                    self.add_log_error("Conversion was not possible")

            else:
                self.add_log_error("Cannot convert:\n  - " + ("\n  - ".join(errors)), True)

        except Exception as e:
            aa.get_python_logger().exception("Conversion failed")
            self.add_log_error("Conversion failed: {}".format(aa.str_exc(e)), True)



import sys
if sys.version_info[0] < 3:
    print("Python version detected:\n*****\n%s\n*****\nCannot run, must be using Python 3" % sys.version)
    sys.exit()

from setuptools import setup, find_packages
from glob import glob

setup(
    name = 'pyfant',
    packages = find_packages(),
    version = '0.16.12.19',
    license = 'GNU GPLv3',
    platforms = 'any',
    description = 'Python API and Graphical applications for PFANT (github.com/trevisanj/PFANT) Spectral Synthesis Software',
    author = 'Julio Trevisan',
    author_email = 'juliotrevisan@gmail.com',
    url = 'https://github.com/trevisanj/pyfant', # use the URL to the github repo
    keywords= ['astronomy', 'spectral synthesis'],
    install_requires = ['numpy', 'scipy', 'astropy', 'matplotlib', 'fortranformat', 'bs4', 'robobrowser',
                        'tabulate', 'astroapi==0.16.12.19'],  # matplotlib never gets installed correctly by pip, but anyway...
    scripts = glob('scripts/*.py')  # Considers system scripts all .py files in 'scripts' directory
)




#!/usr/bin/env python3

"""Converion of molecular lines data to PFANT format"""

import sys
import argparse
import pyfant as pf
import astroapi as aa
import logging
import os


aa.logging_level = logging.INFO
aa.flag_log_file = True


if __name__ == "__main__":

    deffn = pf.FileMolDB.default_filename

    parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=aa.SmartFormatter
    )
    parser.add_argument('fn', type=str, help='Molecular constants database file name',
                        default=deffn, nargs='?')
    args = parser.parse_args()

    if args.fn == deffn and not os.path.isfile(deffn):
        args.fn = None

    m = None
    if args.fn is not None:
        m = pf.FileMolDB()
        m.load(args.fn)
    app = aa.get_QApplication([])
    form = pf.convmol.XConvMol(None, m)
    form.show()
    sys.exit(app.exec_())







